import ast
import error

from ast import Add, BinOp, Constant, Expression, Mult, Name, Sub, expr
from collections.abc import Mapping
from dataclasses import dataclass

type Type = Int


@dataclass
class Int:
    ...


def get_expr_type(e: expr, m: Mapping[str, Type]) -> Type:
    try:
        match e:
            case Name(id):
                if id in m:
                    return m[id]
                else:
                    raise NameError("name '" + id + "' is not defined")
            case Constant(i) if type(i) is int:
                return Int()
            case BinOp(left, Add(), right):
                left_type: Type = get_expr_type(left, m)
                right_type: Type = get_expr_type(right, m)
                match left_type, right_type:
                    case Int(), Int():
                        return Int()
                    case _, _:
                        raise TypeError("unsupported operand type(s) for +")
            case BinOp(left, Sub(), right):
                left_type: Type = get_expr_type(left, m)
                right_type: Type = get_expr_type(right, m)
                match left_type, right_type:
                    case Int(), Int():
                        return Int()
                    case _, _:
                        raise TypeError("unsupported operand type(s) for -")
            case BinOp(left, Mult(), right):
                left_type: Type = get_expr_type(left, m)
                right_type: Type = get_expr_type(right, m)
                match left_type, right_type:
                    case Int(), Int():
                        return Int()
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")
            case _:
                raise TypeError("unsupported expression")
    except Exception as err:
        error.fail(e, err)


def get_type(e: Expression, m: Mapping[str, Type]) -> str:
    match get_expr_type(e.body, m):
        case Int():
            return "int"


src = """
(9 * 2) * (x + 4)
"""


def main() -> None:
    e: Expression = ast.parse(src, mode='eval')
    print(ast.dump(e, indent=4))
    print(src)
    print(get_type(e, {'x': Int()}))


main()
